/* UnLockPRO Version 2.0.0
 * Copyright (C) 1999 Kev Vance <kvance@zeux.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#define USAGE	"USAGE: ulpro <-u|-ns0> <file1> [file2]...\n"\
		"	-u  Unlock the file\n"\
		"	-n  Normal lock the file\n"\
		"	-s  Save lock the file\n"\
		"	-0  S00p3r-l0cK <cough...> the file\n"

#ifdef DOS
#include "getopt.h"
typedef unsigned short int u_int16_t;
#endif

void lowerstring(char *s)
{
	int i;

	for (i = 0; i < strlen(s); i++)
		s[i] = tolower(s[i]);
}

void asciistring(char *s)
{
	int i;

	for (i = 0; i < strlen(s); i++)
		if (s[i] < 32 || s[i] > 126)
			s[i] = '.';
}


int main(int argc, char **argv)
{
	int c, argp, locks, sl_detect, ol_detect;
	u_int16_t boards, bsize, objects, osize;
	FILE *fp;

	char *s;

	int i;
	long position;

	unsigned char buffer[256];
	char *megabuffer;

	/* FLAGS */
	int f_unlock = 0, f_normal = 0, f_save = 0, f_super = 0;

	printf("UnLockPRO 2.0.0 by Kev Vance <kvance@zeux.org>\nLock manipulator for ZZT files\n");
	if (argc == 1) {
		fprintf(stderr, USAGE);
		exit(0);
	}
	while (1) {
		c = getopt(argc, argv, "uns0");
		if (c == -1)
			break;

		switch (c) {
		case 'u':
			if (f_normal || f_save || f_super) {
				fprintf(stderr, "Fatal Error!  Lock or unlock.  Choose one.\n");
				exit(1);
			} else {
				f_unlock = 1;
			}
			break;
		case 'n':
			if (f_unlock) {
				fprintf(stderr, "Fatal Error!  Lock or unlock.  Choose one.\n");
				exit(1);
			} else {
				f_normal = 1;
			}
			break;
		case 's':
			if (f_unlock) {
				fprintf(stderr, "Fatal Error!  Lock or unlock.  Choose one.\n");
				exit(1);
			} else {
				f_save = 1;
			}
			break;
		case '0':
			if (f_unlock) {
				fprintf(stderr, "Fatal Error!  Lock or unlock.  Choose one.\n");
				exit(1);
			} else {
				f_super = 1;
			}
			break;
		case '?':
			fprintf(stderr, "Unrecognized option.");
			fprintf(stderr, USAGE);
			exit(1);
			break;
		}
	}
	if (!f_unlock && !f_normal && !f_super && !f_save) {
		fprintf(stderr, USAGE);
		exit(1);
	}
	if (f_unlock == 1) {
		/* Let's get to work, Shamus! */

		megabuffer = (char *) malloc(65535);
		if (megabuffer == NULL) {
			fprintf(stderr, "Couldn't allocate 64k megabuffer, not checking for omega lock.\n");
		}
		for (argp = 1;; argp++) {
			if (argv[argp][0] != '-')
				break;
			if (argp == argc - 1) {
				fprintf(stderr, "Fatal Error!  I need a filename.\n");
				exit(1);
			}
		}
		for (; argp != argc; argp++) {
			locks = 0;
			sl_detect = 0;
			ol_detect = 0;

			if (argv[argp][0] == '-')
				continue;
			fp = fopen(argv[argp], "r+b");
			if (fp == NULL) {
				fprintf(stderr, "WARNING: Could not open %s... skipping it.\n", argv[argp]);
				continue;
			}
			fread(&buffer, 2, 1, fp);
			if (buffer[0] != 0xff || buffer[1] != 0xff) {
				fprintf(stderr, "WARNING: %s is not a ZZT file... skipping it.\n", argv[argp]);
				fclose(fp);
				continue;
			}
			fseek(fp, 0x02, SEEK_SET);
			fread(&boards, 2, 1, fp);

			fseek(fp, 0x1d, SEEK_SET);
			c = fgetc(fp);
			fread(&buffer, 19, 1, fp);
			if (c > 19)
				c = 19;
			buffer[c] = '\0';
			asciistring((char *) &buffer);
			printf("Loaded world \"%s\" (%d boards)\n", buffer, boards);

			/* Check for normal lock */
			fseek(fp, 0x32, SEEK_SET);
			c = fgetc(fp);
			if (c == 6) {
				printf("  Normal Lock detected...\t");
				fseek(fp, 0x32, SEEK_SET);
				fputc(0, fp);
				printf("nullified.\n");
				locks++;
			}
			/* Check for save lock */
			fseek(fp, 0x108, SEEK_SET);
			c = fgetc(fp);
			if (c != 0) {
				printf("  Save Lock detected...\t\t");
				fseek(fp, 0x108, SEEK_SET);
				fputc(0, fp);
				printf("nullified.\n");
				locks++;
			}
			/* Check for super lock and omega lock */

			bsize = 0x200 - 2;
			fseek(fp, 0, SEEK_SET);

			for (i = 0; i < boards + 1; i++) {
				fseek(fp, bsize + 2, SEEK_CUR);
				position = ftell(fp);
				fread(&bsize, 2, 1, fp);
				/* We'll need that to find the next board later */
				c = fgetc(fp);
				fread(&buffer, 0x20, 1, fp);
				if (c > 20)
					c = 20;
				buffer[c] = '\0';
				/* Check the title for a jump command */
				if (buffer[0] == '!') {
					for (c = 1; c < strlen(buffer); c++) {
						if (buffer[c] == ';') {
							if (sl_detect == 0) {
								printf("  Super Lock detected...\t");
								sl_detect = 1;
								locks++;
							}
							fseek(fp, -33, SEEK_CUR);
							fwrite("\x09UnLockPRO", 10, 1, fp);
							break;
						}
					}
				}
				fseek(fp, position, SEEK_SET);
			}
			if (sl_detect == 1) {
				printf("nullified.\n");
			}
			bsize = 0x200 - 2;
			fseek(fp, 0, SEEK_SET);

			for (i = 0; i < boards + 1; i++) {
				fseek(fp, bsize + 2, SEEK_CUR);
				position = ftell(fp);
				fread(&bsize, 2, 1, fp);
				/* We'll need that to find the next board later */

				fseek(fp, position + 0x35, SEEK_SET);
				/* Run through the RLE */
				for (c = 0; c < 60 * 25;) {
					c += fgetc(fp);
					fread(&buffer, 2, 1, fp);	/* throwaway */
				}
				fseek(fp, 0x56, SEEK_CUR);
				fread(&objects, 2, 1, fp);

				/* Scan objects for "#if debug" */
				for (c = 0; c < objects + 1; c++) {
					if (!megabuffer)
						break;	/* couldn't get the buffer */
					fseek(fp, 0x17, SEEK_CUR);
					fread(&osize, 2, 1, fp);
					fseek(fp, 8, SEEK_CUR);
					if (osize != 0) {
						fread(megabuffer, osize, 1, fp);
						megabuffer[osize] = '\0';
						lowerstring(megabuffer);
						s = strstr(megabuffer, "#if debug");
						if (s) {
							if (ol_detect == 0) {
								printf("Omega Lock detected...\t");
								ol_detect = 1;
								locks++;
							}
							fseek(fp, 0 - osize, SEEK_CUR);
							memcpy(s + 1, "end\r;", 5);
							fwrite(megabuffer, osize, 1, fp);

						}
					}
				}
				fseek(fp, position, SEEK_SET);
			}
			if (ol_detect == 1) {
				printf("nullified.\n");
			}
			printf("---------------------------------------------------------------------------\n");
			printf("TOTALS FOR %s: %d LOCK%sREMOVED\n", argv[argp], locks, locks == 1 ? " " : "S ");
			printf("---------------------------------------------------------------------------\n\n");

			fclose(fp);
		}
		free(megabuffer);
		return 0;
	}
	/* Okay, we're locking */
	for (argp = 1;; argp++) {
		if (argv[argp][0] != '-')
			break;
		if (argp == argc - 1) {
			fprintf(stderr, "Fatal Error!  I need a filename.\n");
			exit(1);
		}
	}
	for (; argp != argc; argp++) {
		locks = 0;
		sl_detect = 0;
		ol_detect = 0;

		if (argv[argp][0] == '-')
			continue;
		fp = fopen(argv[argp], "r+b");
		if (fp == NULL) {
			fprintf(stderr, "WARNING: Could not open %s... skipping it.\n", argv[argp]);
		}
		fread(&buffer, 2, 1, fp);
		if (buffer[0] != 0xff || buffer[1] != 0xff) {
			fprintf(stderr, "WARNING: %s is not a ZZT file... skipping it.\n", argv[argp]);
			fclose(fp);
			continue;
		}
		fseek(fp, 0x02, SEEK_SET);
		fread(&boards, 2, 1, fp);

		fseek(fp, 0x1d, SEEK_SET);
		c = fgetc(fp);
		fread(&buffer, 19, 1, fp);
		if (c > 19)
			c = 19;
		buffer[c] = '\0';
		asciistring((char *) &buffer);
		printf("Loaded world \"%s\" (%d boards)\n", buffer, boards);

		if (f_normal) {
			printf("  Normal locking...\t");
			fseek(fp, 0x32, SEEK_SET);
			fputc(6, fp);
			printf("done.\n");
		}
		if (f_save) {
			printf("  Save locking...\t");
			fseek(fp, 0x108, SEEK_SET);
			fputc(1, fp);
			printf("done.\n");
		}
		if (boards == 0 && f_super) {
			fprintf(stderr, "ERROR: You need at least 2 boards to super lock.\n");
			f_super = 0;
		}
		if (f_super) {
			printf("  Super locking...\t");

			bsize = 0x200 - 2;
			fseek(fp, 0, SEEK_SET);

			for (i = 0; i < boards + 1; i++) {
				fseek(fp, bsize + 2, SEEK_CUR);
				position = ftell(fp);
				if (i == boards) {
					fwrite("\x04\x00\x02:u", 5, 1, fp);
				} else {
					fread(&bsize, 2, 1, fp);
				}
				/* We'll need that to find the next board later */
				fwrite("\x1c!u;UnLockPRO brand SUPERLOCK", 29, 1, fp);
				fseek(fp, position, SEEK_SET);
			}
			printf("done.\n");
		}
		fclose(fp);
	}
	return 0;
}
